@@ -66,10 +66,18 @@ extern int imap_before_smtp;
  */
 int _ic_capability(struct ImapSession *self)
 {
+	field_t val;
+	gboolean override = FALSE;
+
 	if (!check_state_and_args(self, "CAPABILITY", 0, 0, -1))
 		return 1;	/* error, return */
 
-	dbmail_imap_session_printf(self, "* CAPABILITY %s\r\n", IMAP_CAPABILITY_STRING);
+	
+	GETCONFIGVALUE("capability", "IMAP", val);
+	if (strlen(val) > 0)
+		override = TRUE;
+
+	dbmail_imap_session_printf(self, "* CAPABILITY %s\r\n", override ? val : IMAP_CAPABILITY_STRING);
 	dbmail_imap_session_printf(self, "%s OK CAPABILITY completed\r\n", self->tag);
 
 	return 0;
@@ -148,6 +156,7 @@ int _ic_login(struct ImapSession *self)
 	child_reg_connected_user(self->args[self->args_idx]);
 
 	dbmail_imap_session_printf(self, "%s OK LOGIN completed\r\n", self->tag);
+
 	return 0;
 }
 
@@ -208,6 +217,8 @@ int _ic_authenticate(struct ImapSession *self)
 
 	dbmail_imap_session_printf(self, "%s OK AUTHENTICATE completed\r\n", self->tag);
 	
+	child_reg_connected_user(username);
+
 	g_free(username);
 	g_free(password);
 	return 0;
@@ -243,6 +254,8 @@ int _ic_select(struct ImapSession *self)
 	if ((result = dbmail_imap_session_mailbox_open(self, mailbox))) 
 		return result;
 
+	dbmail_imap_session_set_state(self,IMAPCS_SELECTED);
+
 	if ((result = dbmail_imap_session_mailbox_show_info(self)))
 		return result;
 
@@ -271,7 +284,6 @@ int _ic_select(struct ImapSession *self)
 		return -1;
 	}
 
-	dbmail_imap_session_set_state(self,IMAPCS_SELECTED);
 	dbmail_imap_session_printf(self, "%s OK [%s] SELECT completed\r\n", self->tag,
 		permstring);
 	return 0;
@@ -297,14 +309,14 @@ int _ic_examine(struct ImapSession *self)
 	if ((result = dbmail_imap_session_mailbox_open(self, mailbox)))
 		return result;
 
+	/* update permission: examine forces read-only */
+	ud->mailbox.permission = IMAPPERM_READ;
+
+	dbmail_imap_session_set_state(self,IMAPCS_SELECTED);
+
 	if ((result = dbmail_imap_session_mailbox_show_info(self)))
 		return result;
 	
-	/* update permission: examine forces read-only */
-	ud->mailbox.permission = IMAPPERM_READ;
-	
-	dbmail_imap_session_set_state(self,IMAPCS_SELECTED);
-
 	dbmail_imap_session_printf(self, "%s OK [READ-ONLY] EXAMINE completed\r\n", self->tag);
 	return 0;
 }
@@ -641,7 +653,7 @@ int _ic_unsubscribe(struct ImapSession *self)
 		return 1;
 
 	if (! (mboxid = dbmail_imap_session_mailbox_get_idnr(self, self->args[0]))) {
-		dbmail_imap_session_printf(self, "%s NO mailbox does not exist\r\n", self->tag);
+		dbmail_imap_session_printf(self, "%s OK UNSUBSCRIBE on mailbox that does not exist\r\n", self->tag);
 		return 0;
 	}
 
@@ -892,6 +904,27 @@ int _ic_status(struct ImapSession *self)
 	return 0;
 }
 
+/*
+ * _ic_idle
+ *
+ * non-expunging close for select mailbox and return to AUTH state
+ *
+ */
+
+int _ic_idle(struct ImapSession *self)
+{
+	int result;
+	if (!check_state_and_args(self, "IDLE", 0, 0, IMAPCS_AUTHENTICATED))
+		return 1;	/* error, return */
+
+	if ((result = dbmail_imap_session_idle(self)) != 0)
+		return result;
+
+	dbmail_imap_session_printf(self, "%s OK IDLE terminated\r\n", self->tag);
+	return 0;
+}
+
+
 
 /*
  * _ic_append()
@@ -1355,7 +1388,8 @@ int _ic_thread(struct ImapSession *self)
 	if (MATCH(self->args[0],"ORDEREDSUBJECT"))
 		return sorted_search(self,SEARCH_THREAD_ORDEREDSUBJECT);
 	if (MATCH(self->args[0],"REFERENCES"))
-		return sorted_search(self,SEARCH_THREAD_REFERENCES);
+		dbmail_imap_session_printf(self, "%s BAD THREAD=REFERENCES not supported\r\n",self->tag);
+		//return sorted_search(self,SEARCH_THREAD_REFERENCES);
 
 	return 1;
 }
@@ -1374,7 +1408,7 @@ int _dm_imapsession_get_ids(struct ImapSession *self, const char *set)
 	while (retry > 0) {
 
 		self->ids = dbmail_mailbox_get_set(self->mailbox, set, self->use_uid);
-		if (g_tree_nnodes(self->ids)> 0)
+		if (self->ids && g_tree_nnodes(self->ids)> 0)
 			break;
 
 		if ((result = dbmail_mailbox_open(self->mailbox))!= DM_SUCCESS)
@@ -1382,7 +1416,7 @@ int _dm_imapsession_get_ids(struct ImapSession *self, const char *set)
 		retry--;
 	}
 
-	if (g_tree_nnodes(self->ids)==0) {
+	if ( (!self->ids) || (g_tree_nnodes(self->ids)==0) ) {
 		dbmail_imap_session_printf(self, "%s BAD invalid message range specified\r\n", self->tag);
 		return DM_EGENERAL;
 	}
