@@ -45,6 +45,7 @@ int reallyquiet = 0;
 char *configFile = DEFAULT_CONFIG_FILE;
 
 int has_errors = 0;
+int serious_errors = 0;
 
 /* set up database login data */
 extern db_param_t _db_params;
@@ -55,6 +56,7 @@ static int do_check_integrity(void);
 static int do_null_messages(void);
 static int do_purge_deleted(void);
 static int do_set_deleted(void);
+static int do_dangling_aliases(void);
 static int do_header_cache(void);
 static int do_check_iplog(const char *timespec);
 static int do_check_replycache(const char *timespec);
@@ -68,13 +70,14 @@ int do_showhelp(void) {
 //	0........10........20........30........40........50........60........70........80
 	"Use this program to maintain your DBMail database.\n"
 	"See the man page for more info. Summary:\n\n"
-	"     -a        perform all checks (in this release: -ctubpd)\n"
+	"     -a        perform all checks (in this release: -ctubpds)\n"
 	"     -c        clean up database (optimize/vacuum)\n"
 	"     -t        test for message integrity\n"
 	"     -u        null message check\n"
 	"     -b        body/header/envelope cache check\n"
 	"     -p        purge messages have the DELETE status set\n"
 	"     -d        set DELETE status for deleted messages\n"
+	"     -s        remove dangling/invalid aliases and forwards\n"
 	"     -r time   clear the replycache used for autoreply/vacation\n"
 	"     -l time   clear the IP log used for IMAP/POP-before-SMTP\n"
 	"               the time syntax is [<hours>h][<minutes>m]\n"
@@ -99,7 +102,7 @@ int main(int argc, char *argv[])
 	int check_iplog = 0, check_replycache = 0;
 	char *timespec_iplog = NULL, *timespec_replycache = NULL;
 	int null_messages = 0;
-	int vacuum_db = 0, purge_deleted = 0, set_deleted = 0;
+	int vacuum_db = 0, purge_deleted = 0, set_deleted = 0, dangling_aliases = 0;
 	int show_help = 0;
 	int do_nothing = 1;
 	int is_header = 0;
@@ -111,7 +114,7 @@ int main(int argc, char *argv[])
 
 	/* get options */
 	opterr = 0;		/* suppress error message from getopt() */
-	while ((opt = getopt(argc, argv, "-acbtl:r:pud" "i" "f:qnyvVh")) != -1) {
+	while ((opt = getopt(argc, argv, "-acbtl:r:puds" "i" "f:qnyvVh")) != -1) {
 		/* The initial "-" of optstring allows unaccompanied
 		 * options and reports them as the optarg to opt 1 (not '1') */
 		switch (opt) {
@@ -120,6 +123,7 @@ int main(int argc, char *argv[])
 			vacuum_db = 1;
 			purge_deleted = 1;
 			set_deleted = 1;
+			dangling_aliases = 1;
 			check_integrity = 1;
 			null_messages = 1;
 			is_header = 1;
@@ -146,6 +150,11 @@ int main(int argc, char *argv[])
 			do_nothing = 0;
 			break;
 
+		case 's':
+			dangling_aliases = 1;
+			do_nothing = 0;
+			break;
+
 		case 't':
 			check_integrity = 1;
 			do_nothing = 0;
@@ -227,7 +236,7 @@ int main(int argc, char *argv[])
 
  	/* Don't make any changes unless specifically authorized. */
  	if (!yes_to_all) {
-		qprintf("Choosing dry-run mode. Use -y to disable it.");
+		qprintf("Choosing dry-run mode. No changes will be made at this time.\n");
 		no_to_all = 1;
  	}
 
@@ -235,42 +244,52 @@ int main(int argc, char *argv[])
 	SetTraceLevel("DBMAIL");
 	GetDBParams(&_db_params);
 
-	qprintf("Opening connection to database... \n");
+	qverbosef("Opening connection to database... \n");
 	if (db_connect() != 0) {
 		qerrorf("Failed. An error occured. Please check log.\n");
 		return -1;
 	}
 
-	qprintf("Opening connection to authentication... \n");
+	qverbosef("Opening connection to authentication... \n");
 	if (auth_connect() != 0) {
 		qerrorf("Failed. An error occured. Please check log.\n");
 		return -1;
 	}
 
-	qprintf("Ok. Connected.\n");
+	qverbosef("Ok. Connected.\n");
 
 	if (check_integrity) do_check_integrity();
 	if (null_messages) do_null_messages();
 	if (purge_deleted) do_purge_deleted();
 	if (is_header) do_header_cache();
 	if (set_deleted) do_set_deleted();
+	if (dangling_aliases) do_dangling_aliases();
 	if (check_iplog) do_check_iplog(timespec_iplog);
 	if (check_replycache) do_check_replycache(timespec_replycache);
 	if (vacuum_db) do_vacuum_db();
 
-	if (!has_errors) {
+	if (!has_errors && !serious_errors) {
 		qprintf("\nMaintenance done. No errors found.\n");
 	} else {
 		qerrorf("\nMaintenance done. Errors were found");
-		if (no_to_all) {
+		if (serious_errors) {
+			qerrorf(" but not fixed due to failures.\n");
+			qerrorf("Please check the logs for further details, "
+				"turning up the trace level as needed.\n");
+			// Indicate that something went really wrong
+			has_errors = 3;
+		} else if (no_to_all) {
 			qerrorf(" but not fixed.\n");
-			qerrorf("Try running dbmail-util with the '-y' "
-			    "option to repair the errors.\n");
-		}
-		if (yes_to_all) {
+			qerrorf("Run again with the '-y' option to "
+				"repair the errors.\n");
+			// Indicate that the program should be run with -y
+			has_errors = 2;
+		} else if (yes_to_all) {
 			qerrorf(" and fixed.\n");
-			qerrorf("Try running dbmail-util again to confirm "
-			    "that the errors were repaired.\n");
+			qerrorf("We suggest running dbmail-util again to "
+				"confirm that all errors were repaired.\n");
+			// Indicate that the program should be run again
+			has_errors = 1;
 		}
 	}
 
@@ -279,7 +298,7 @@ int main(int argc, char *argv[])
 	config_free();
 	g_mime_shutdown();
 	
-	return 0;
+	return has_errors;
 }
 
 int do_purge_deleted(void)
@@ -291,6 +310,7 @@ int do_purge_deleted(void)
 		if (db_deleted_count(&deleted_messages) < 0) {
 			qerrorf
 			    ("Failed. An error occured. Please check log.\n");
+			serious_errors = 1;
 			return -1;
 		}
 		qprintf("Ok. [%llu] messages have DELETE status.\n",
@@ -301,6 +321,7 @@ int do_purge_deleted(void)
 		if (db_deleted_purge(&deleted_messages) < 0) {
 			qerrorf
 			    ("Failed. An error occured. Please check log.\n");
+			serious_errors = 1;
 			return -1;
 		}
 		qprintf("Ok. [%llu] messages deleted.\n", deleted_messages);
@@ -318,6 +339,7 @@ int do_set_deleted(void)
 		if (db_count_deleted(&messages_set_to_delete) == -1) {
 			qerrorf
 			    ("Failed. An error occured. Please check log.\n");
+			serious_errors = 1;
 			return -1;
 		}
 		qprintf("Ok. [%llu] messages need to be set for deletion.\n",
@@ -328,6 +350,7 @@ int do_set_deleted(void)
 		if (db_set_deleted(&messages_set_to_delete) == -1) {
 			qerrorf
 			    ("Failed. An error occured. Please check log.\n");
+			serious_errors = 1;
 			return -1;
 		}
 		qprintf("Ok. [%llu] messages set for deletion.\n",
@@ -336,6 +359,7 @@ int do_set_deleted(void)
 		if (db_calculate_quotum_all() < 0) {
 			qerrorf
 			    ("Failed. An error occured. Please check log.\n");
+			serious_errors = 1;
 			return -1;
 		}
 		qprintf("Ok. Used quota updated for all users.\n");
@@ -343,6 +367,100 @@ int do_set_deleted(void)
 	return 0;
 }
 
+GList *find_dangling_aliases(const char * const name)
+{
+	int result;
+	char *username;
+	struct dm_list uids;
+	struct dm_list fwds;
+	GList *userids = NULL;
+	GList *dangling = NULL;
+
+	/* For each alias, figure out if it resolves to a valid user
+	 * or some forwarding address. If neither, remove it. */
+
+	dm_list_init(&fwds);
+	dm_list_init(&uids);
+	result = auth_check_user_ext(name,&uids,&fwds,0);
+	
+	if (!result) {
+		qerrorf("Nothing found searching for [%s].\n", name);
+		serious_errors = 1;
+		return dangling;
+	}
+
+	if (dm_list_getstart(&uids))
+		userids = g_list_copy_list(userids,dm_list_getstart(&uids));
+
+	userids = g_list_first(userids);
+	while (userids) {
+		username = auth_get_userid(*(u64_t *)userids->data);
+		if (!username) {
+			dangling = g_list_prepend(dangling, userids->data);
+		}
+		g_free(username);
+		if (! g_list_next(userids))
+			break;
+		userids = g_list_next(userids);
+	}
+
+	return dangling;
+}
+
+int do_dangling_aliases(void)
+{
+	int count = 0;
+	int result = 0;
+	GList *aliases = NULL;
+
+	if (no_to_all)
+		qprintf("\nCounting aliases with nonexistent delivery userid's...\n");
+	if (yes_to_all)
+		qprintf("\nRemoving aliases with nonexistent delivery userid's...\n");
+
+	aliases = auth_get_known_aliases();
+	aliases = g_list_dedup(aliases);
+	aliases = g_list_first(aliases);
+	while (aliases) {
+		char deliver_to[21];
+		GList *dangling = find_dangling_aliases(aliases->data);
+
+		dangling = g_list_first(dangling);
+		while (dangling) {
+			count++;
+			g_snprintf(deliver_to, 21, "%llu", *(u64_t *)dangling->data);
+			qverbosef("Dangling alias [%s] delivers to nonexistent user [%s]\n",
+				(char *)aliases->data, deliver_to);
+			if (yes_to_all) {
+				if (auth_removealias_ext(aliases->data, deliver_to) < 0) {
+					qerrorf("Error: could not remove alias [%s] deliver to [%s] \n",
+						(char *)aliases->data, deliver_to);
+					serious_errors = 1;
+					result = -1;
+				}
+			}
+			if (!g_list_next(dangling))
+				break;
+			dangling = g_list_next(dangling);
+		}
+		g_list_destroy(g_list_first(dangling));
+
+		if (! g_list_next(aliases))
+			break;
+		aliases = g_list_next(aliases);
+	}
+	g_list_destroy(g_list_first(aliases));
+
+	if (count > 0) {
+		qerrorf("Ok. Found [%d] dangling aliases.\n", count);
+		has_errors = 1;
+	} else {
+		qprintf("Ok. Found [%d] dangling aliases.\n", count);
+	}
+
+	return result;
+}
+
 int do_null_messages(void)
 {
 	time_t start, stop;
@@ -360,6 +478,7 @@ int do_null_messages(void)
 
 	if (db_icheck_null_messages(&lostlist) < 0) {
 		qerrorf("Failed. An error occured. Please check log.\n");
+		serious_errors = 1;
 		return -1;
 	}
 
@@ -395,6 +514,7 @@ int do_null_messages(void)
 	time(&start);
 	if (db_icheck_null_physmessages(&lostlist) < 0) {
 		qerrorf("Failed, an error occured. Please check log.\n");
+		serious_errors = 1;
 		return -1;
 	}
 
@@ -429,14 +549,17 @@ int do_check_integrity(void)
 	time_t start, stop;
 	struct dm_list lostlist;
 	struct element *el;
+	const char *action;
+	int count = 0;
 	u64_t id;
 
-	if (no_to_all) {
-		qprintf("\nChecking DBMAIL messageblocks integrity...\n");
-	}
-	if (yes_to_all) {
-		qprintf("\nRepairing DBMAIL messageblocks integrity...\n");
-	}
+	if (yes_to_all)
+		action = "Repairing";
+	else
+		action = "Checking";
+
+	qprintf("\n%s DBMAIL messageblocks integrity...\n", action);
+
 	time(&start);
 
 	/* This is what we do:
@@ -449,6 +572,7 @@ int do_check_integrity(void)
 	/* first part */
 	if (db_icheck_messageblks(&lostlist) < 0) {
 		qerrorf("Failed. An error occured. Please check log.\n");
+		serious_errors = 1;
 		return -1;
 	}
 
@@ -486,18 +610,44 @@ int do_check_integrity(void)
 
 
 	time(&stop);
-	qverbosef("--- checking block integrity took %g seconds\n",
-	       difftime(stop, start));
-	qverbosef("--- checking block integrity took %g seconds\n",
-		difftime(stop, start));
+	qverbosef("--- %s block integrity took %g seconds\n",
+	       action, difftime(stop, start));
+
 
 	/* second part */
 	start = stop;
-	qprintf("\nChecking DBMAIL message integrity...\n");
+	qprintf("\n%s DBMAIL physmessage integrity...\n", action);
+	if ((count = db_icheck_physmessages(FALSE)) < 0) {
+		qerrorf("Failed. An error occurred. Please check log.\n");
+		serious_errors = 1;
+		return -1;
+	}
+	if (count > 0) {
+		qerrorf("Ok. Found [%d] unconnected physmessages", count);
+
+		if (yes_to_all) {
+			if (db_icheck_physmessages(TRUE) < 0)
+				qerrorf("Warning: could not delete orphaned physmessages. Check log.\n");
+			else
+				qerrorf("Ok. Orphaned physmessages deleted.\n");
+		}
+	} else {
+		qprintf("Ok. Found [%d] unconnected physmessages.\n", count);
+	}
+
+	time(&stop);
+	qverbosef("--- %s unconnected physmessages took %g seconds\n",
+		action, difftime(stop, start));
+
+
+	/* third part */
+	start = stop;
+	qprintf("\n%s DBMAIL message integrity...\n", action);
 
 	if (db_icheck_messages(&lostlist) < 0) {
 		qerrorf
 		    ("Failed. An error occured. Please check log.\n");
+		serious_errors = 1;
 		return -1;
 	}
 
@@ -539,18 +689,17 @@ int do_check_integrity(void)
 	}
 
 	time(&stop);
-	qverbosef("--- checking message integrity took %g seconds\n",
-	       difftime(stop, start));
-	qverbosef("--- checking message integrity took %g seconds\n",
-		difftime(stop, start));
+	qverbosef("--- %s message integrity took %g seconds\n",
+	       action, difftime(stop, start));
 
-	/* third part */
-	qprintf("\nChecking DBMAIL mailbox integrity...\n");
+	/* fourth part */
+	qprintf("\n%s DBMAIL mailbox integrity...\n", action);
 	start = stop;
 
 	if (db_icheck_mailboxes(&lostlist) < 0) {
 		qerrorf
 		    ("Failed. An error occured. Please check log.\n");
+		serious_errors = 1;
 		return -1;
 	}
 
@@ -586,10 +735,8 @@ int do_check_integrity(void)
 	}
 
 	time(&stop);
-	qverbosef("--- checking mailbox integrity took %g seconds\n",
-	       difftime(stop, start));
-	qverbosef("--- checking mailbox integrity took %g seconds\n",
-		difftime(stop, start));
+	qverbosef("--- %s mailbox integrity took %g seconds\n",
+	       action, difftime(stop, start));
 	
 	return 0;
 }
@@ -609,6 +756,7 @@ static int do_is_header(void)
 
 	if (db_icheck_isheader(&lost) < 0) {
 		qerrorf("Failed. An error occured. Please check log.\n");
+		serious_errors = 1;
 		return -1;
 	}
 
@@ -650,6 +798,7 @@ static int do_rfc_size(void)
 
 	if (db_icheck_rfcsize(&lost) < 0) {
 		qerrorf("Failed. An error occured. Please check log.\n");
+		serious_errors = 1;
 		return -1;
 	}
 
@@ -692,6 +841,7 @@ static int do_envelope(void)
 
 	if (db_icheck_envelope(&lost) < 0) {
 		qerrorf("Failed. An error occured. Please check log.\n");
+		serious_errors = 1;
 		return -1;
 	}
 
@@ -725,14 +875,20 @@ int do_header_cache(void)
 	time_t start, stop;
 	GList *lost = NULL;
 	
-	if (do_rfc_size())
+	if (do_rfc_size()) {
+		serious_errors = 1;
 		return -1;
+	}
 
-	if (do_is_header())
+	if (do_is_header()) {
+		serious_errors = 1;
 		return -1;
+	}
 	
-	if (do_envelope())
+	if (do_envelope()) {
+		serious_errors = 1;
 		return -1;
+	}
 	
 	if (no_to_all) 
 		qprintf("\nChecking DBMAIL for cached header values...\n");
@@ -743,6 +899,7 @@ int do_header_cache(void)
 
 	if (db_icheck_headercache(&lost) < 0) {
 		qerrorf("Failed. An error occured. Please check log.\n");
+		serious_errors = 1;
 		return -1;
 	}
 
@@ -756,7 +913,7 @@ int do_header_cache(void)
 	if (yes_to_all) {
 		if (db_set_headercache(lost) < 0) {
 			qerrorf("Error caching the header values ");
-			has_errors = 1;
+			serious_errors = 1;
 		}
 	}
 
@@ -778,6 +935,7 @@ int do_check_iplog(const char *timespec)
 	if (find_time(timespec, &timestring) != 0) {
 		qerrorf("\nFailed to find a timestring: [%s] is not <hours>h<minutes>m.\n",
 		       timespec);
+		serious_errors = 1;
 		return -1;
 	}
 
@@ -785,6 +943,7 @@ int do_check_iplog(const char *timespec)
 		qprintf("\nCounting IP entries older than [%s]...\n", timestring);
 		if (db_count_iplog(timestring, &log_count) < 0) {
 			qerrorf("Failed. An error occured. Check the log.\n");
+			serious_errors = 1;
 			return -1;
 		}
 		qprintf("Ok. [%llu] IP entries are older than [%s].\n",
@@ -794,6 +953,7 @@ int do_check_iplog(const char *timespec)
 		qprintf("\nRemoving IP entries older than [%s]...\n", timestring);
 		if (db_cleanup_iplog(timestring, &log_count) < 0) {
 			qerrorf("Failed. Please check the log.\n");
+			serious_errors = 1;
 			return -1;
 		}
 
@@ -811,6 +971,7 @@ int do_check_replycache(const char *timespec)
 	if (find_time(timespec, &timestring) != 0) {
 		qerrorf("\nFailed to find a timestring: [%s] is not <hours>h<minutes>m.\n",
 		       timespec);
+		serious_errors = 1;
 		return -1;
 	}
 
@@ -818,6 +979,7 @@ int do_check_replycache(const char *timespec)
 		qprintf("\nCounting RC entries older than [%s]...\n", timestring);
 		if (db_count_replycache(timestring, &log_count) < 0) {
 			qerrorf("Failed. An error occured. Check the log.\n");
+			serious_errors = 1;
 			return -1;
 		}
 		qprintf("Ok. [%llu] RC entries are older than [%s].\n",
@@ -827,6 +989,7 @@ int do_check_replycache(const char *timespec)
 		qprintf("\nRemoving RC entries older than [%s]...\n", timestring);
 		if (db_cleanup_replycache(timestring, &log_count) < 0) {
 			qerrorf("Failed. Please check the log.\n");
+			serious_errors = 1;
 			return -1;
 		}
 
@@ -846,6 +1009,7 @@ int do_vacuum_db(void)
 		fflush(stdout);
 		if (db_cleanup() < 0) {
 			qerrorf("Failed. Please check the log.\n");
+			serious_errors = 1;
 			return -1;
 		}
 
@@ -871,16 +1035,22 @@ int find_time(const char *timespec, timestring_t *timestring)
 
 	time(&td);		/* get time */
 
-	if (!timespec)
+	if (!timespec) {
+		serious_errors = 1;
 		return -1;
+	}
 
 	/* find first num */
 	tmp = strtol(timespec, &end, 10);
-	if (!end)
+	if (!end) {
+		serious_errors = 1;
 		return -1;
+	}
 
-	if (tmp < 0)
+	if (tmp < 0) {
+		serious_errors = 1;
 		return -1;
+	}
 
 	switch (*end) {
 	case 'h':
@@ -892,12 +1062,15 @@ int find_time(const char *timespec, timestring_t *timestring)
 	case 'M':
 		hour = 0;
 		min = tmp;
-		if (end[1])	/* should end here */
+		if (end[1]) {	/* should end here */
+			serious_errors = 1;
 			return -1;
+		}
 
 		break;
 
 	default:
+		serious_errors = 1;
 		return -1;
 	}
 
@@ -906,14 +1079,20 @@ int find_time(const char *timespec, timestring_t *timestring)
 	if (timespec[end - timespec + 1]) {
 		tmp = strtol(&timespec[end - timespec + 1], &end, 10);
 		if (end) {
-			if ((*end != 'm' && *end != 'M') || end[1])
+			if ((*end != 'm' && *end != 'M') || end[1]) {
+				serious_errors = 1;
 				return -1;
+			}
 
-			if (tmp < 0)
+			if (tmp < 0) {
+				serious_errors = 1;
 				return -1;
+			}
 
-			if (min >= 0)	/* already specified minutes */
+			if (min >= 0) {	/* already specified minutes */
+				serious_errors = 1;
 				return -1;
+			}
 
 			min = tmp;
 		}
