@@ -52,6 +52,8 @@ const int month_len[] = { 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
 
 extern const char *imap_flag_desc_escaped[];
 
+static struct DbmailIconv *ic;
+
 #undef max
 #define max(x,y) ( (x) > (y) ? (x) : (y) )
 
@@ -99,7 +101,7 @@ void create_unique_id(char *target, u64_t message_idnr)
 	char *md5_str;
 
 	a_message_idnr = g_strdup_printf("%llu",message_idnr);
-	a_rand = g_strdup_printf("%d",rand());
+	a_rand = g_strdup_printf("%d",g_random_int());
 
 	if (message_idnr != 0)
 		snprintf(target, UID_SIZE, "%s:%s",
@@ -537,22 +539,15 @@ char * dm_strnesc(const char * from, size_t len)
 }
 	
 /* 
- *
- * replace tabs with spaces and all multi-spaces with single spaces 
- *
+ * replace all multi-spaces with single spaces 
  */
-
-void  dm_pack_spaces(char *in) 
+static void pack_char(char *in, char c)
 {
-	char *tmp, *saved;
-	/* replace tabs with spaces */
-	g_strdelimit(in,"\t",' ');
-	
-	/* replace all multi-spaces with single spaces */
-	tmp = g_strdup(in);
+	char *saved;
+	char *tmp = g_strdup(in);
 	saved = tmp;
 	while(*tmp) {
-		if ((*tmp == ' ') && (*(tmp+1) == ' ')) {
+		if ((*tmp == c) && (*(tmp+1) == c)) {
 			tmp++;
 		} else {
 			*in++=*tmp++;
@@ -561,6 +556,19 @@ void  dm_pack_spaces(char *in)
 	g_free(saved);
 	*in='\0';
 }
+
+/* 
+ *
+ * replace tabs with spaces and all multi-spaces with single spaces 
+ *
+ */
+
+void  dm_pack_spaces(char *in) 
+{
+	/* replace tabs with spaces */
+	g_strdelimit(in,"\t",' ');
+	pack_char(in,' ');
+}
 /* 
  * base-subject
  *
@@ -645,7 +653,7 @@ char * dm_base_subject(const char *subject)
 	if (g_mime_utils_text_is_8bit((unsigned char *)subject, strlen(subject))) 
 		tmp = g_strdup(subject);
 	else 
-		tmp = g_mime_utils_header_decode_text((unsigned char *)subject);
+		tmp = dbmail_iconv_decode_text(subject);
 	saved = tmp;
 	
 	dm_pack_spaces(tmp);
@@ -1497,7 +1505,7 @@ static void _structure_part_message_rfc822(GMimeObject *part, gpointer data, gbo
 
 static void get_param_list(gpointer key, gpointer value, gpointer data)
 {
-	gchar *s = g_mime_utils_header_encode_text((unsigned char *)((GMimeParam *)value)->value);
+	gchar *s = g_mime_utils_header_encode_text((const char *)((GMimeParam *)value)->value);
 	*(GList **)data = g_list_append_printf(*(GList **)data, "\"%s\"", (char *)key);
 	*(GList **)data = g_list_append_printf(*(GList **)data, "\"%s\"", s);
 	g_free(s);
@@ -1889,7 +1897,7 @@ GList* dbmail_imap_append_alist_as_plist(GList *list, const InternetAddressList
 
 			/* personal name */
 			if (ia->name && ia->value.addr) {
-				name = g_mime_utils_header_encode_phrase((unsigned char *)ia->name);
+				name = g_mime_utils_header_encode_phrase((const char *)ia->name);
 				g_strdelimit(name,"\"\\",' ');
 				g_strstrip(name);
 				s = dbmail_imap_astring_as_string(name);
@@ -2015,7 +2023,7 @@ static GList * envelope_address_part(GList *list, GMimeMessage *message, const c
 	result = g_mime_message_get_header(message,header);
 	
 	if (result) {
-		result_enc = convert_8bit_field_to_utf8(result, charset);
+		result_enc = dbmail_iconv_str_to_utf8(result, charset);
 		t = imap_cleanup_address(result_enc);
 		g_free(result_enc);
 		alist = internet_address_parse_string(t);
@@ -2063,58 +2071,87 @@ char * message_get_charset(GMimeMessage *message)
 }
 
 
+
+void dbmail_iconv_init(void)
+{
+	static gboolean initialized = FALSE;
+	iconv_t tmp_i = (iconv_t)-1;
+
+	if (initialized)
+		return;
+
+	ic = g_new0(struct DbmailIconv,1);	
+
+	memset(ic->db_charset,'\0', FIELDSIZE);
+	memset(ic->msg_charset,'\0', FIELDSIZE);
+
+	ic->to_db = (iconv_t)-1;
+	ic->from_msg = (iconv_t)-1;
+
+	GETCONFIGVALUE("ENCODING", "DBMAIL", ic->db_charset);
+	GETCONFIGVALUE("DEFAULT_MSG_ENCODING", "DBMAIL", ic->msg_charset);
+
+	if(ic->db_charset[0]) {
+		if ((tmp_i=g_mime_iconv_open(ic->db_charset,"UTF-8")) == (iconv_t)-1) {
+			g_strlcpy(ic->db_charset, g_mime_locale_charset(),FIELDSIZE);
+		} else {
+			g_mime_iconv_close(tmp_i);
+		}
+	} else {
+		g_strlcpy(ic->db_charset,g_mime_locale_charset(), FIELDSIZE);
+	}
+
+	if (ic->msg_charset[0]) {
+		if ((tmp_i = g_mime_iconv_open(ic->msg_charset,"UTF-8")) == (iconv_t)-1) {
+			g_strlcpy(ic->msg_charset, g_mime_locale_charset(), FIELDSIZE);
+
+		} else {
+			g_mime_iconv_close(tmp_i);
+		}
+	} else {
+		g_strlcpy(ic->msg_charset, g_mime_locale_charset(), FIELDSIZE);
+	}
+
+
+	TRACE(TRACE_DEBUG,"Initialize DB encoding surface [UTF-8..%s]", ic->db_charset);
+	ic->to_db = g_mime_iconv_open(ic->db_charset,"UTF-8");
+	assert(ic->to_db != (iconv_t)-1);
+
+	TRACE(TRACE_DEBUG,"Initialize DB decoding surface [%s..UTF-8]", ic->db_charset);
+	ic->from_db = g_mime_iconv_open("UTF-8", ic->db_charset);
+	assert(ic->to_db != (iconv_t)-1);
+
+	TRACE(TRACE_DEBUG,"Initialize default MSG decoding surface [%s..UTF-8]", ic->msg_charset);
+	ic->from_msg=g_mime_iconv_open("UTF-8",ic->msg_charset);
+	assert(ic->from_msg != (iconv_t)-1);
+	
+	initialized = TRUE;
+}
+
 /* convert not encoded field to utf8 */
-char * convert_8bit_field_to_utf8(const char* str_in, const char *charset)
+char * dbmail_iconv_str_to_utf8(const char* str_in, const char *charset)
 {
 	char * subj=NULL;
-	static iconv_t default_iconv=(iconv_t)-1;
-	static int allocated_default_iconv = 0;
-	iconv_t conv_iconv;
-	
-	// Init static vars
-	if (default_iconv==(iconv_t)-1){
-		field_t val;
-		const char *default_charset=NULL;
-		GETCONFIGVALUE("DEFAULT_MSG_ENCODING", "DBMAIL", val);
-		if (val[0])
-			default_charset=val;
-		else
-			default_charset="x-unknown";// use locale charset
-		default_iconv=g_mime_iconv_open("UTF-8",default_charset);
-		if (default_iconv == (iconv_t)-1) {
-			TRACE(TRACE_DEBUG,"incorrect default encoding [%s]", default_charset);
-		}else
-			allocated_default_iconv = 1;
-	}
-	// End init
-	
-	if (str_in==NULL) {
-		return NULL;
-	}
+	iconv_t conv_iconv = (iconv_t)-1;
 
-	if (g_utf8_validate((const gchar *)str_in, -1, NULL) || !g_mime_utils_text_is_8bit((unsigned char *)str_in, strlen(str_in))) {
-		// Conversion not needed
+	dbmail_iconv_init();
+
+	if (str_in==NULL)
+		return NULL;
+
+	if (g_utf8_validate((const gchar *)str_in, -1, NULL) || !g_mime_utils_text_is_8bit((unsigned char *)str_in, strlen(str_in)))
 		return g_strdup(str_in);
-	}
 
  	if (charset) {
- 		// codepage not set in message header use default
- 		TRACE(TRACE_DEBUG,"encoding 8bit use charset [%s]", charset);
-  			
- 		if ((conv_iconv=g_mime_iconv_open("UTF-8",charset))==(iconv_t)-1) {
-			TRACE(TRACE_DEBUG,"incorrect encoding [%s] base [UTF-8]", charset);
-			subj=g_mime_iconv_strdup(default_iconv,str_in);
-		} else {
+ 		if ((conv_iconv=g_mime_iconv_open("UTF-8",charset)) != (iconv_t)-1) {
 			subj=g_mime_iconv_strdup(conv_iconv,str_in);
 			g_mime_iconv_close(conv_iconv);
 		}
 	}
-	if (subj==NULL) {
-		subj=g_mime_iconv_strdup(default_iconv,str_in);
-	}
+	if (subj==NULL)
+		subj=g_mime_iconv_strdup(ic->from_msg,str_in);
 	    
 	if (subj==NULL) {
-		// On Error convertion,  replace all 8 bit symbol with '?'
 		subj=g_strdup(str_in);
 		char *p;
 		for(p=subj;*p;p++)
@@ -2124,81 +2161,39 @@ char * convert_8bit_field_to_utf8(const char* str_in, const char *charset)
 	return subj;
 }
 
-
 /* convert not encoded field to database encoding */
-char * convert_8bit_field(const char* str_in, const char *charset)
+char * dbmail_iconv_str_to_db(const char* str_in, const char *charset)
 {
 	char * subj=NULL;
-	static const char * base_charset=NULL;
-	static iconv_t base_iconv=(iconv_t)-1;
-	static iconv_t default_iconv=(iconv_t)-1;
 	iconv_t conv_iconv;
 
-	// Init static vars
-	if(base_charset==NULL){
-		field_t val;
-		GETCONFIGVALUE("ENCODING", "DBMAIL", val);
-		if(val[0]) {
-			base_charset=val;
-			TRACE(TRACE_DEBUG,"Base charset [%s]", base_charset);
-			iconv_t tmp_i=g_mime_iconv_open(base_charset,"UTF-8");
-			if(tmp_i == (iconv_t)-1) {
-				base_charset=g_mime_locale_charset();
-				TRACE(TRACE_DEBUG,"Base charset test filed set to [%s]", base_charset);
-			}
-			else
-				g_mime_iconv_close(tmp_i);
-		}
-		else
-		        base_charset=g_mime_locale_charset();
-		base_charset=g_strdup(base_charset);
-		base_iconv=g_mime_iconv_open(base_charset,base_charset);
-		if (base_iconv == (iconv_t)-1)
-			TRACE(TRACE_DEBUG,"incorrect base encoding [%s]", base_charset);
-			
-	}
-	// End init static
-	
-	if (base_iconv == (iconv_t)-1) {
-		const char *default_charset=NULL;
-		field_t val;
-		GETCONFIGVALUE("DEFAULT_MSG_ENCODING", "DBMAIL", val);
-		if (val[0])
-			default_charset=val;
-		else
-			default_charset="x-unknown";// use locale charset
-		default_iconv=g_mime_iconv_open(base_charset,default_charset);
-		if (default_iconv == (iconv_t)-1) {
-			TRACE(TRACE_DEBUG,"incorrect default encoding [%s]", default_charset);
-		}
-	}
-	if (str_in==NULL)
+	dbmail_iconv_init();
+
+	if ( str_in == NULL )
 		return NULL;
 	
-	if (!g_mime_utils_text_is_8bit((unsigned char *)str_in, strlen(str_in))) {
-		// No convertion required
+	if (! g_mime_utils_text_is_8bit((unsigned char *)str_in, strlen(str_in)) )
 		return g_strdup(str_in);
-	}
 
-	if ((subj=g_mime_iconv_strdup(base_iconv,str_in))!=NULL) {
-		// Conversion already done by header decode ? May insert to database
+	if ((subj=g_mime_iconv_strdup(ic->to_db,str_in)) != NULL)
 		return subj;
-	}
 
  	if (charset) {
- 	        TRACE(TRACE_DEBUG,"encoding 8bit use charset [%s]", charset);
-  			
- 		if ((conv_iconv=g_mime_iconv_open(base_charset,charset))==(iconv_t)-1) {
- 			TRACE(TRACE_WARNING,"incorrect encoding [%s] base [%s]", charset,base_charset);
- 			subj=g_mime_iconv_strdup(default_iconv,str_in);
- 		} else {
+ 		if ((conv_iconv=g_mime_iconv_open(ic->db_charset,charset)) != (iconv_t)-1) {
  			subj=g_mime_iconv_strdup(conv_iconv,str_in);
  			g_mime_iconv_close(conv_iconv);
   		}
   	}
     
 	if (subj==NULL) {
-		// On Error convertion,  replace all 8 bit symbol with '?'
+		char *subj2;
+		if ((subj2 = g_mime_iconv_strdup(ic->from_msg,str_in)) != NULL) {
+			subj = g_mime_iconv_strdup(ic->to_db, subj2);
+			g_free(subj2);
+		}
+	}
+
+	if (subj==NULL) {
 		subj=g_strdup(str_in);
 		char *p;
 		for(p=subj;*p;p++)
@@ -2208,48 +2203,147 @@ char * convert_8bit_field(const char* str_in, const char *charset)
 	return subj;
 }
 /* encode string from database encoding to mime (7-bit) */
-char * convert_8bit_db_to_mime(const char* str_in)
+char * dbmail_iconv_db_to_utf7(const char* str_in)
 {
 	char * subj=NULL;
-	static iconv_t base_iconv=(iconv_t)-1;
-	
-	if (base_iconv == (iconv_t)-1) { //Init
-		field_t val;
-		const char * base_charset=NULL;
-		GETCONFIGVALUE("ENCODING", "DBMAIL", val);
-		if(val[0]) {
-			base_charset=val;
-			TRACE(TRACE_DEBUG,"Base charset [%s]", base_charset);
-			iconv_t tmp_i=g_mime_iconv_open("UTF-8", base_charset);
-			if(tmp_i == (iconv_t)-1) {
-				base_charset=g_mime_locale_charset();
-				TRACE(TRACE_DEBUG,"Base charset test filed set to [%s]", base_charset);
-			}
-			else
-				g_mime_iconv_close(tmp_i);
-		}
-		else
-		        base_charset=g_mime_locale_charset();
-		base_charset=g_strdup(base_charset);
-		base_iconv=g_mime_iconv_open("UTF-8", base_charset);
-		if (base_iconv == (iconv_t)-1)
-			TRACE(TRACE_DEBUG,"incorrect base encoding [%s]", base_charset);
-	}
 	
+	dbmail_iconv_init();
+
 	if (str_in==NULL)
 		return NULL;
 	
 	if (!g_mime_utils_text_is_8bit((unsigned char *)str_in, strlen(str_in)))
 		return g_strdup(str_in);
 
-	if ((! g_utf8_validate((const char *)str_in,-1,NULL)) && ((subj=g_mime_iconv_strdup(base_iconv,str_in))!=NULL)) {
+	if ((! g_utf8_validate((const char *)str_in,-1,NULL)) && ((subj=g_mime_iconv_strdup(ic->from_db, str_in))!=NULL)) {
  		gchar *subj2;
-		subj2 = g_mime_utils_header_encode_text((unsigned char *)subj);
+		subj2 = g_mime_utils_header_encode_text((const char *)subj);
   		g_free(subj);
  		return subj2;
   	}
 
-	return g_mime_utils_header_encode_text((unsigned char *)str_in);
+	return g_mime_utils_header_encode_text(str_in);
+}
+
+/* work around a bug in gmime (< 2.2.10) where utf7 strings are not completely decoded */
+char * dbmail_iconv_decode_text(const char *in)
+{
+	size_t i=0, l=0, r=0, len, wlen=0;
+	char p2=0, p = 0, c, n = 0;
+	char *res, *s;
+	gboolean inchar = FALSE, inword = FALSE;
+	GString *buf = g_string_new("");
+	GString *str = g_string_new("");
+
+	len = strlen(in);
+	for (i = 0; i<len; i++)
+	{
+		c = in[i];
+		n = in[i+1];
+
+		if ((c == '=') && (n == '?') && (inword == FALSE) && (inchar == FALSE)) {
+			inchar = TRUE;
+			l = i;
+		} else if (((p2 == 'q') || (p2 == 'Q') || (p2 == 'b') || (p2 == 'B')) && (p == '?') && inchar) {
+			inchar = FALSE;
+			inword = TRUE;
+			wlen = 0;
+		} else if ((p2 == '?') && (p == '=') && inword && wlen) {
+			inword = FALSE;
+			r = i;
+		} else if (inword) {
+			wlen++;
+		}
+
+		if (l < r) {
+			s = g_mime_utils_header_decode_text(buf->str);
+			g_string_append_printf(str, "%s", s);
+			g_free(s);
+			l = r;
+			i = r;
+			g_string_printf(buf,"%s","");
+			g_string_append_c(str, in[i]);
+		} else if (inword || inchar) {
+			g_string_append_c(buf, in[i]);
+		} else {
+			g_string_append_c(str, in[i]);
+		}
+
+		p2 = p;
+		p = c;
+	}
+	if (buf->len) {
+		s = g_mime_utils_header_decode_text(buf->str);
+		g_string_append_printf(str, "%s", s);
+		g_free(s);
+	}
+	g_string_free(buf,TRUE);
+
+	res = str->str;
+	g_string_free(str,FALSE);
+
+	return res;
+
+}
+/*
+ * dbmail_iconv_decode_address
+ * \param address the raw address header value
+ * \result allocated string containing a utf8 encoded representation
+ * 		of the input address
+ *
+ * 	paranoia rulez here, since input is untrusted
+ */
+char * dbmail_iconv_decode_address(char *address)
+{
+	InternetAddressList *l;
+	char *r = NULL, *t = NULL, *s = NULL;
+
+	if (address == NULL) return NULL;
+
+ 	// first make the address rfc2047 compliant if it happens to 
+ 	// contain 8bit data
+	if ( (g_mime_utils_text_is_8bit((unsigned char *)address, strlen(address))) )
+		s = g_mime_utils_header_encode_text((const char *)address);
+	else
+		s = g_strdup(address);
+
+	// cleanup broken addresses before parsing
+	t = imap_cleanup_address((const char *)s); g_free(s);
+
+	// feed the result to gmime's address parser and
+	// render it back to a hopefully now sane string
+	l = internet_address_parse_string(t); g_free(t);
+	s = internet_address_list_to_string(l, FALSE);
+	internet_address_list_destroy(l);
+
+	// now we're set to decode the rfc2047 address header
+	// into clean utf8
+	r = dbmail_iconv_decode_text(s); g_free(s);
+
+	// oops: if the rfc2047 encoded address names contain
+	// encoded double-quotes, imap_cleanup_address has added quotes that are now
+	// redundant. we need to strip those
+	pack_char(r,'"');
+
+	return r;
+}
+
+char * dbmail_iconv_decode_field(const char *in, const char *charset, gboolean isaddr)
+{
+	char *tmp_raw;
+	char *value;
+
+	if ((tmp_raw = dbmail_iconv_str_to_utf8((const char *)in, charset)) == NULL) {
+		TRACE(TRACE_WARNING, "unable to decode headervalue [%s] using charset [%s]", in, charset);
+		return NULL;
+	}
+	
+	if (isaddr)
+		value = dbmail_iconv_decode_address(tmp_raw);
+	else
+		value = dbmail_iconv_decode_text(tmp_raw);
+
+	return value;
 }
 
 
@@ -2282,10 +2376,9 @@ char * imap_get_envelope(GMimeMessage *message)
 
 	if (result) {
 		char *charset = message_get_charset(message);
-		char * subj = convert_8bit_field_to_utf8(result, charset);
+		char * subj = dbmail_iconv_str_to_utf8(result, charset);
 		g_free(charset);
-		s = g_mime_utils_header_encode_text((unsigned char *)subj);
-		TRACE(TRACE_DEBUG,"encoding 8bit subject [%s] -> [%s]", subj, s);
+		s = g_mime_utils_header_encode_text((const char *)subj);
 		t = dbmail_imap_astring_as_string(s);
 		g_free(s);
 		g_free(subj);
@@ -2431,7 +2524,7 @@ char * imap_cleanup_address(const char *a)
 {
 	char *r, *t;
 	char *inptr;
-	char prev,next=0;
+	char prev, next=0;
 	unsigned incode=0, inquote=0;
 	size_t i, l;
 	GString *s;
@@ -2447,50 +2540,46 @@ char * imap_cleanup_address(const char *a)
 	dm_pack_spaces(t);
 	inptr = t;
 	inptr = g_strstrip(inptr);
-	prev = *inptr;
-	if (*(inptr+1))
-		next=*(inptr+1);
+	prev = inptr[0];
 	
-	// quote encoded string
-	if (*inptr == '=' && next=='?') {
-		if (prev != '"')
-			g_string_append_c(s,'"');
-		inquote=1;
-		incode=1;
-	}
-
 	l = strlen(inptr);
+
 	for (i = 0; i < l - 1; i++) {
 
-		next=inptr[i+1];
-
-		if (inptr[i] == '<')
-			break;
-
-		if (inptr[i] == ' ' && next=='<')
-			break;
-
-		if (inptr[i] == '=' && next=='?')
-			incode=1;
-
-		if (prev=='?' && inptr[i]=='=')
-			incode=0;
+		next = inptr[i+1];
 
 		if (incode && (inptr[i] == '"' || inptr[i] == ' '))
 			continue; // skip illegal chars inquote
 
-		prev = inptr[i];
+		if ((! inquote) && inptr[i]=='"')
+			inquote = 1;
+		else if (inquote && inptr[i] == '"')
+			inquote = 0;
+
+		// quote encoded string
+		if (inptr[i] == '=' && next == '?' && (! incode)) {
+			incode=1;
+			if (prev != '"' && (! inquote)) {
+				g_string_append_c(s,'"');
+				inquote = 1;
+			}
+		} 
 
 		g_string_append_c(s,inptr[i]); 
 
+		if (inquote && incode && prev == '?' && inptr[i] == '=' && (next == '"' || next == ' ' || next == '<')) {
+			if ((next != '"' ) && ((i < l-2) && (inptr[i+2] != '='))) {
+				g_string_append_c(s, '"');
+				inquote = 0;
+			}
+			if (next == '<')
+				g_string_append_c(s,' ');
+			incode=0;
+		}
+
+		prev = inptr[i];
 	}
 
-	if (inquote)
-		g_string_append_c(s,'"');
-
-	if (inptr[i] == '<' && prev != inptr[i] && prev != ' ')
-		g_string_append_c(s,' ');
-		
 	inptr+=i;
 
 	if (*inptr)
@@ -2540,4 +2629,17 @@ char * imap_flags_as_string(msginfo_t *msginfo)
 	return s;
 }
 
+long long unsigned dm_strtoull(const char *nptr, char **endptr, int base)
+{
+	errno = 0;
+	long long int r = strtoll(nptr, endptr, base);
+	if (errno)
+		return (long long unsigned)0;
+
+	if (r < 0) {
+		errno = EINVAL;
+		return (long long unsigned)0;
+	}
+	return (long long unsigned)r;
+}
 
