@@ -88,6 +88,26 @@ GList * auth_get_known_users(void)
 	return users;
 }
 
+GList * auth_get_known_aliases(void)
+{
+	u64_t i;
+	GList * aliases = NULL;
+
+	snprintf(__auth_query_data, AUTH_QUERY_SIZE,
+		 "SELECT alias FROM %saliases ORDER BY alias",DBPFX);
+
+	if (__auth_query(__auth_query_data) == -1) {
+		TRACE(TRACE_ERROR, "could not retrieve user list");
+		return NULL;
+	}
+
+	for (i = 0; i < (unsigned) db_num_rows(); i++) 
+		aliases = g_list_append(aliases, g_strdup(db_get_result(i, 0)));
+	
+	db_free_result();
+	return aliases;
+}
+
 int auth_getclientid(u64_t user_idnr, u64_t * client_idnr)
 {
 	const char *query_result;
@@ -186,7 +206,6 @@ int auth_check_user_ext(const char *username, struct dm_list *userids, struct dm
 	int occurences = 0;
 	void *saveres;
 	u64_t counter;
-	const char *query_result;
 	char *endptr;
 	u64_t id;
 	unsigned num_rows;
@@ -256,9 +275,10 @@ int auth_check_user_ext(const char *username, struct dm_list *userids, struct dm
 	if (num_rows > 0) {
 		for (counter = 0; counter < num_rows; counter++) {
 			/* do a recursive search for deliver_to */
-			query_result = db_get_result(counter, 0);
+			char *query_result = g_strdup(db_get_result(counter, 0));
 			TRACE(TRACE_DEBUG, "checking user %s to %s", username, query_result);
 			occurences += auth_check_user_ext(query_result, userids, fwds, checks+1 );
+			g_free(query_result);
 		}
 	}
 	db_free_result();
